from .debug import debug_print
from .encoding import (
    DESC_ENCODING,
    ENCODING,
    MISS_ENCODING,
    TARG_ENCODING,
    change_role,
    code_to_query,
    encode_attribute,
    get_att,
    get_att_2d,
    query_to_code,
)
from .decoration import (
    DecoratedDecisionTreeClassifier,
    DecoratedDecisionTreeRegressor,
    DecoratedRandomForestRegressor,
    DecoratedRandomForestClassifier,
)

from .data_handling import(get_i_o)

from .test_data import load_iris, default_dataset
